/**
 * Gaming Tools
 * Copyright (C) 2003  Michael Birken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Michael Birken (o__1@hotmail.com)
 * @version 2.1
 */
package gamingtools;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/**
 * A class that desires to generate animation frames extends this
 * class. The abstract methods are invoked in the following order:
 * <OL>
 * <LI><CODE>init()</CODE> -- Called once prior to frame generation on a
 * different thread than the one for the animation loop. It is used to load
 * images and sounds, and for configuration.
 * <LI><CODE>updateState()</CODE> -- Used to update the state of the game.
 * It will occasionally be invoked more often than the render methods to
 * maintain the apparent frame rate.
 * <LI><CODE>isBackgroundSame()</CODE> -- Returns <CODE>true</CODE> when the
 * background has not changed. It is used to determine which
 * <CODE>renderBackground()</CODE> method to call.
 * <LI><CODE>renderBackground()</CODE> -- Two overloads, the first draws the
 * complete background, the second a sub-region of it. If the background has
 * not changed and the contents of the buffers have not been lost, then
 * the latter overload is used repeatedly to restore the dirty regions created
 * by sprite drawing. Otherwise, the former version is used.
 * <LI><CODE>renderForeground()</CODE> -- Draws the sprites and other
 * foreground graphics. It should use <CODE>drawSprite()</CODE> and
 * <CODE>markDirtyRegion()</CODE> appropriately.
 * </OL>
 * Avoid creating temporary objects in these methods because the incremental
 * garbage collection that occurs to clean up those objects will introduce
 * noticeable pauses.
 * @see gamingtools.GraphicsSource
 */
public abstract class FrameBuilder implements RenderListener {

  /**
   * Rectangular dirty region.
   */
  protected class Region {
    /** Horizontal position of the upper-left corner.*/
    public int x;
    /** Vertical position of the upper-left corner. */
    public int y;
    /** Width of the region. */
    public int width;
    /** Height of the region. */
    public int height;
  }

  /**
   * Description of an animation frame.
   */
  protected class FrameInfo {
    /** An index that uniquely identifies the background. */
    int backgroundNumber;
    /** Number of dirty regions associated with this frame. */
    int dirtyRegionCount;
  }

  /** Default maximum dirty regions stored in the history per buffer. */
  public static final int DEFAULT_MAX_DIRTY_REGIONS = 64;
  /** Maximum number of dirty regions stored in the history per buffer. */
  public final int MAX_DIRTY_REGIONS;
  /** Equal to the number of buffers. */
  protected int historySize;
  /** Stored dirty regions: <CODE>[history index][region index]</CODE> */
  protected Region[][] regionHistory;
  /** Stored frame descriptions. */
  protected FrameInfo[] frameInfoHistory;
  /** Alternates between 0 and 1, or remains at 0. */
  protected int historyIndex;
  /** Index that is incremented only if the background changes. */
  protected int backgroundNumber;
  /** Width of the buffers. */
  protected int bufferWidth;
  /** Height of the buffers. */
  protected int bufferHeight;

  /**
   * Constructs this class with dirty region history size equal to the
   * specified value.
   * @param maxDirtyRegions Maximum number of dirty regions that can be
   *   recorded for each buffer.
   */
  public FrameBuilder(int maxDirtyRegions) {
    MAX_DIRTY_REGIONS = maxDirtyRegions;
  }

  /**
   * Constructs this class with a dirty region history size equal to 64. If
   * more than 64 dirty regions are required, use the alternate constructor.
   */
  public FrameBuilder() {
    this(DEFAULT_MAX_DIRTY_REGIONS);
  }

  /**
   * <P>This method is automatically invoked by <CODE>GraphicsSource</CODE>. Do
   * not call it directly.</P>
   * <P>This method initializes the dirty region history and delegates the
   * call to the abstract method <CODE>init()</CODE>.</P>
   * @param isPageFlipping Indicates whether the <CODE>BufferStrategy</CODE>
   *   within <CODE>GraphicsSource</CODE> employs page flipping. If
   *   <CODE>true</CODE>, then even frames and odd frames are drawn to different
   *   offscreen buffers. Otherwise, all frames are drawn to the same buffer.
   *   This parameter determines how the buffer must be updated to generate the
   *   next frame.
   * @param bufferWidth The width of the buffers.
   * @param bufferHeight The height of the bufrers.
   * @see #init()
   */
  public final void init(
      boolean isPageFlipping, int bufferWidth, int bufferHeight) {

    historySize = isPageFlipping ? 2 : 1;
    regionHistory = new Region[historySize][MAX_DIRTY_REGIONS];
    frameInfoHistory = new FrameInfo[historySize];

    for(int i = historySize - 1; i >= 0; i--) {
      frameInfoHistory[i] = new FrameInfo();
      for(int j = MAX_DIRTY_REGIONS - 1; j >= 0; j--)
        regionHistory[i][j] = new Region();
    }

    this.bufferWidth = bufferWidth;
    this.bufferHeight = bufferHeight;

    init();
  }

  /**
   * <P>This method is automatically invoked by <CODE>GraphicsSource</CODE>. Do
   * not call it directly.</P>
   * <P>This method invokes the abstract method <CODE>updateState()</CODE>
   * followed by the abstract method <CODE>isBackgroundSame()</CODE>.</P>
   * @see #updateState()
   * @see #isBackgroundSame()
   */
  public final void updateModel() {
    updateState();
    if (!isBackgroundSame())
      backgroundNumber++;
  }

  /**
   * <P>This method is automatically invoked by <CODE>GraphicsSource</CODE>. Do
   * not call it directly.</P>
   * <P>This method invokes one of the abstract <CODE>renderBackground()</CODE>
   * methods followed by the abstract <CODE>renderForeground()</CODE>
   * method.</P>
   * @param g The graphics context used to draw the animation frame. Disposal of
   *   this graphics context is automatically performed by
   *   <CODE>GraphicsSource</CODE>.
   * @param isBufferCleared Indicates whether the buffer is in the same
   *   state since the last time it was drawn to. Occationally, the buffer
   *   may be losted and restored. When this happens, it is effectively erased.
   *   If this parameter is <CODE>true</CODE>, then the entire buffer needs to
   *   be painted. Otherwise, only the portion of the buffer that changes for
   *   this frame needs to be updated. Note, after calling
   *   <CODE>GraphicsSource.setRenderListener()</CODE>, this parameter will
   *   be <CODE>true</CODE> for the first frame or two depending upon which
   *   buffer strategy is employed.
   * @param bufferIndex If page flipping is used, then this parameter will
   *   alternate between 0 and 1. Otherwise, it will always be 0.
   * @see #renderBackground(Graphics)
   * @see #renderBackground(Graphics,int,int,int,int)
   * @see #renderForeground(Graphics)
   */
  public final void render(
      Graphics g, boolean isBufferCleared, int bufferIndex) {

    if (isBufferCleared)
      backgroundNumber++;

    historyIndex = bufferIndex;

    FrameInfo frameInfo = frameInfoHistory[historyIndex];
    if (frameInfo.backgroundNumber == backgroundNumber) {
      Region[] regions = regionHistory[historyIndex];
      for(int i = frameInfo.dirtyRegionCount - 1; i >= 0; i--) {
        Region region = regions[i];
        g.setClip(region.x, region.y, region.width, region.height);
        renderBackground(g, region.x, region.y, region.width, region.height);
      }
    } else
      renderBackground(g);

    frameInfo.backgroundNumber = backgroundNumber;
    frameInfo.dirtyRegionCount = 0;

    g.setClip(0, 0, bufferWidth, bufferHeight);
    renderForeground(g);
  }

  /**
   * Draws an image at the specified coordinates and marks the region it
   * occupies as dirty.
   * @param g A graphics context to the back buffer.
   * @param image Sprite to draw.
   * @param x Horizontal position of the upper-left corner.
   * @param y Vertical position of the upper-left corner.
   * @see #markDirtyRegion(int,int,int,int)
   */
  public void drawSprite(Graphics g, Image image, int x, int y) {
    markDirtyRegion(x, y, image.getWidth(null), image.getHeight(null));
    g.drawImage(image, x, y, null);
  }

  /**
   * Records a dirty region. The region will be restored in a successive frame
   * by invoking the sub-region form of <CODE>drawBackground()</CODE>.
   * @param x Horizontal position of the upper-left corner.
   * @param y Vertical position of the upper-left corner.
   * @param width Width of the region.
   * @param height Height of the region.
   * @see #drawSprite(Graphics,Image,int,int)
   */
  public void markDirtyRegion(int x, int y, int width, int height) {

    if (x > bufferWidth || y > bufferHeight || x + width < 0 || y + height < 0)
      return;

    if (x < 0) {
      width += x;
      x = 0;
    }

    if (y < 0) {
      height += y;
      y = 0;
    }

    if (x + width > bufferWidth)
      width = bufferWidth - x;

    if (y + height > bufferHeight)
      height = bufferHeight - y;

    FrameInfo frameInfo = frameInfoHistory[historyIndex];
    Region region = regionHistory[historyIndex][frameInfo.dirtyRegionCount];
    region.x = x;
    region.y = y;
    region.width = width;
    region.height = height;
    frameInfo.dirtyRegionCount++;
  }

  /**
   * This method is used to load images and sounds, and for configuration. It
   * is invoked prior to frame generation on a different thread than the one
   * for the animation loop.
   */
  public abstract void init();

  /**
   * This method is used to update the state of the game. To sustain the
   * apparent frame rate, this method will occasionally be invoked more often
   * than the render methods.
   */
  public abstract void updateState();

  /**
   * Returns <CODE>true</CODE> if and only if the background has not changed.
   * It is used to determine which <CODE>renderBackground()</CODE> method to
   * call. This method is invoked as often as <CODE>updateState()</CODE>, which
   * is typically more often than the render methods because of frame skipping.
   * @return <CODE>true</CODE> when the background has not changed;
   * <CODE>false</CODE> otherwise.
   */
  public abstract boolean isBackgroundSame();

  /**
   * Draws the complete background. Do not invoke <CODE>g.dispose()</CODE>
   * since <CODE>GraphicsSource</CODE> already takes care of that.
   * @param g A graphics context to the back buffer.
   */
  public abstract void renderBackground(Graphics g);

  /**
   * Draws a rectangular sub-region of the background. This method is intended
   * to restore a dirty region of the background created when a sprite or other
   * foreground graphic was drawn to this buffer in a previous frame. Do not
   * invoke <CODE>g.dispose()</CODE> since <CODE>GraphicsSource</CODE> already
   * takes care of that. <CODE>g</CODE> has a clipping region set to the
   * dirty region.
   * @param g A graphics context to the back buffer.
   * @param x Horizontal position of the upper-left corner.
   * @param y Vertical position of the upper-left corner.
   * @param width Width of the region.
   * @param height Heigh of the region.
   */
  public abstract void renderBackground(
      Graphics g, int x, int y, int width, int height);

  /**
   * Draws the sprites and other foreground graphics. It should use
   * <CODE>markDirtyRegion()</CODE>, which records a dirty region for
   * restoration in a successive frame, and <CODE>drawSprite()</CODE>, which is
   * a convenience method that invokes <CODE>g.drawImage()</CODE> followed by
   * <CODE>markDirtyRegion()</CODE>.
   * @param g A graphics context to the back buffer.
   */
  public abstract void renderForeground(Graphics g);
}