/**
 * Gaming Tools
 * Copyright (C) 2003  Michael Birken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Michael Birken (o__1@hotmail.com)
 * @version 2.1
 */
package gamingtools;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.BufferCapabilities.FlipContents;

/**
 * The <CODE>GraphicsSource</CODE> class provides methods for entering and
 * exiting full-screen exclusive mode and for registering for animation
 * frame rendering callbacks. All of the methods of this class are static.
 */
public final class GraphicsSource {

  /**
   * Used when the application runs in a window instead of full-screen exclusive
   * mode. This listener exits the program.
   */
  public static final WindowAdapter DEFAULT_WINDOW_LISTENER
      = new WindowAdapter() {

    /**
     * Invoked when the window is in the process of being closed.
     * @param e Event associated with that process.
     */
    public void windowClosing(WindowEvent e) {
      exitDisplayMode();
      System.exit(NORMAL_EXIT);
    }
  };

  /**
   * The default RenderListener used when no RenderListener is set. This
   * RenderListener simply paints the buffers black.
   */
  private static final RenderListener DEFAULT_RENDER_LISTENER
      = new RenderListener() {

    /** Width of the buffers. */
    private int bufferWidth;
    /** Height of the buffers. */
    private int bufferHeight;

    /**
     * Invoked by GraphicsSource prior to frame generation. Captures
     * the buffer dimensions.
     * @param isPageFlipping Type of the buffer strategy.
     * @param bufferWidth Width of the buffers.
     * @param bufferHeight Height of the buffers.
     */
    public void init(
        boolean isPageFlipping, int bufferWidth, int bufferHeight) {
      this.bufferWidth = bufferWidth;
      this.bufferHeight = bufferHeight;
    }

    /**
     * Since there is no model to update, this methods does nothing.
     */
    public void updateModel() {
    }

    /**
     * Fills the back buffer with black.
     * @param g Graphics context to the back buffer.
     * @param isBufferCleared Buffer contents state.
     * @param bufferIndex Index of the back buffer.
     */
    public void render(Graphics g, boolean isBufferCleared, int bufferIndex) {
      if (isBufferCleared) {
        g.setColor(Color.BLACK);
        g.fillRect(0, 0, bufferWidth, bufferHeight);
      }
    }
  };

  /**
   * Used when no exception handler is provided. This exception handler
   * displays the stack trace to standard out and shuts down the application.
   */
  private static final ExceptionHandler DEFAULT_EXCEPTION_HANDLER
      = new ExceptionHandler() {

    /**
     * Called back an <CODE>Exception</CODE> or an <CODE>Error</CODE> happens.
     * @param t The <CODE>Exception</CODE> or <CODE>Error</CODE>.
     */
    public void exceptionOccurred(Throwable t) {
      t.printStackTrace();
      exitDisplayMode();
      System.exit(EXCEPTION_EXIT);
    }
  };

  /**
   * Number of frames with a delay of 0 ms before the animation thread yields
   * to other running threads.
   */
  private static final int NO_DELAYS_PER_YIELD = 16;
  /** This return code indicates that the program shutdown normally. */
  private static final int NORMAL_EXIT = 0;
  /** This return code indicates that the program shutdown improperly. */
  private static final int EXCEPTION_EXIT = 1;
  /** Used as the full screen window. */
  private static Frame frame;
  /** Describes the graphics devices available in this environment. */
  private static GraphicsDevice graphicsDevice;
  /** Provides double buffering for fluid animation. */
  private static BufferStrategy bufferStrategy;
  /** Regularly called back to generate successive animation frames. */
  private static volatile RenderListener renderListener;
  /** true if and only if in full screen mode. */
  private static boolean isFullScreenMode;
  /** true if and only if page flipping is used as the buffer strategy. */
  private static boolean isPageFlipping;
  /** Thread used to execute the animation loop. */
  private static Thread animationThread;
  /** true when the animation loop is being executed. */
  private static volatile boolean animating;
  /** Exception handler used when an exception occurs in the animation loop. */
  private static ExceptionHandler exceptionHandler = DEFAULT_EXCEPTION_HANDLER;
  /** Width of the buffers. */
  private static int bufferWidth;
  /** Height of the buffers. */
  private static int bufferHeight;
  /** Display mode prior to changing it. */
  private static DisplayMode originalDisplayMode;
  /** true if enterDisplayMode() executed without an error. */
  private static boolean enterDisplayModeSuccessful;
  /** Alternates between 0 and 1 for page flipping; otherwise, always 0. */
  private static int bufferIndex;
  /** Set to 1 or 2 after calling setRenderListener() to force clearing. */
  private static volatile int clearBufferCount;

  /**
   * <P>Sets the <CODE>RenderListener</CODE> used for generating animation
   * frames.</P>
   * <P>This method invokes <CODE>init()</CODE> on a different thread than the
   * calling thread. This enables <CODE>init()</CODE> to perform blocking
   * calls, such as loading images and sounds, while an animation continues
   * to run. After <CODE>init()</CODE> completes, <CODE>GraphicsSource</CODE>
   * switches over to the specified <CODE>RenderListener</CODE>.</P>
   * @param renderListener <CODE>RenderListener</CODE> used for generating
   *   animation frames. Or, <CODE>null</CODE> to switch back to the default
   *   <CODE>RenderListener</CODE>. The default <CODE>RenderListener</CODE>
   *   simply paints the buffers completely black.
   * @see #clearRenderListener()
   */
  public static void setRenderListener(final RenderListener renderListener) {
    checkDisplayMode();
    Thread thread = new Thread(new Runnable() {
      public void run() {
        try {
          RenderListener oldListener = GraphicsSource.renderListener;
          RenderListener newListener = (renderListener == null)
              ? DEFAULT_RENDER_LISTENER : renderListener;

          newListener.init(isPageFlipping, bufferWidth, bufferHeight);
          GraphicsSource.renderListener = newListener;
          if (isPageFlipping)
            clearBufferCount = 2;
          else
            clearBufferCount = 1;
        } catch(Throwable t) {
          exceptionHandler.exceptionOccurred(t);
        }
      }
    });
    thread.setDaemon(true);
    thread.start();
  }

  /**
   * Unregisters the <CODE>RenderListener</CODE> if there is one. This method is
   * equivalent to passing <CODE>null</CODE> to
   * <CODE>setRenderListener()</CODE>.
   * @see #setRenderListener(RenderListener)
   */
  public static void clearRenderListener() {
    checkDisplayMode();
    setRenderListener(DEFAULT_RENDER_LISTENER);
  }

  /**
   * Starts the animation loop.
   * @param framesPerSecond The apparent frame rate in fps.
   */
  public synchronized static void startAnimation(final int framesPerSecond) {
    startAnimation(framesPerSecond, framesPerSecond);
  }

  /**
   * Starts the animation loop.
   * @param framesPerSecond The apparent frame rate in fps.
   * @param maxFrameSkips Maximum number of frames to skip. Use 0 to indicate
   *   no frame skipping.
   */
  public synchronized static void startAnimation(
      final int framesPerSecond, final int maxFrameSkips) {

    checkDisplayMode();

    if (bufferStrategy == null)
      return;

    final long FRAME_PERIOD = 1000000000L / framesPerSecond;

    if (animationThread != null)
      return;

    animationThread = new Thread() {
      public void run() {

        try {
          RenderListener listener = DEFAULT_RENDER_LISTENER;
          int clearCount = 0;
          long fractions = 0L;
          long overSleepTime = 0L;
          int noDelays = 0;
          animating = true;

          long nextFrameStart = System.nanoTime();
          while(animating) {

            if (clearBufferCount > 0) {
              clearCount = clearBufferCount;
              clearBufferCount = 0;
              listener = renderListener;
            }
            
            long remaining = nextFrameStart - System.nanoTime();
            if (remaining > 0) {
              try {
                Thread.sleep(remaining / 1000000);
              } catch(Throwable t) {          
              }
            }

            do {
              listener.updateModel();
              nextFrameStart += FRAME_PERIOD;
            } while(nextFrameStart < System.nanoTime());              

            Graphics g = bufferStrategy.getDrawGraphics();

            if (bufferStrategy.contentsLost())
              clearCount = (isPageFlipping) ? 2 : 1;
            else {
              boolean isBufferCleared = false;
              if (clearCount > 0) {
                clearCount--;
                isBufferCleared = true;
              }
              listener.render(g, isBufferCleared, bufferIndex);
              bufferStrategy.show();
              if (isPageFlipping)
                bufferIndex = (bufferIndex == 0) ? 1 : 0;
            }
            g.dispose();
          }
        } catch(Throwable t) {
          exceptionHandler.exceptionOccurred(t);
        }
      }
    };

    animationThread.setDaemon(true);
    animationThread.start();
  }

  /**
   * Stops the animation loop. This method does not return until the animation
   * loop thread terminates.
   */
  public synchronized static void stopAnimation() {
    if (!animating || animationThread == null)
      return;

    animating = false;
    try {
      animationThread.join();
    } catch(InterruptedException e) {
      throw new RuntimeException(e);
    }
    animationThread = null;
  }

  /**
   * <P>Attempts to enter full-screen exclusive mode with the specified
   * screen dimensions. Of all available display modes, it enters the one with
   * the highest bit depth and refresh rate. Bit depth is considered a higher
   * priority.</P>
   * <P>If full-screen exclusive mode is unobtainable, it generates a
   * <CODE>Frame</CODE> with a visible drawing surface of the specified size
   * and the program will run outside of full-screen exclusive mode.</P>
   * @param width Width of full screen size.
   * @param height Height of full screen size.
   * @see #getFrame()
   * @see #DEFAULT_WINDOW_LISTENER
   */
  public static synchronized void enterDisplayMode(int width, int height) {

    if (frame != null)
      return;

    bufferWidth = width;
    bufferHeight = height;

    try {
      GraphicsEnvironment graphicsEnvironment
          = GraphicsEnvironment.getLocalGraphicsEnvironment();
      graphicsDevice = graphicsEnvironment.getDefaultScreenDevice();
      GraphicsConfiguration graphicsConfiguration
          = graphicsDevice.getDefaultConfiguration();
      originalDisplayMode = graphicsDevice.getDisplayMode();
      DisplayMode displayMode = getDisplayMode(width, height);

      frame = new Frame(graphicsConfiguration);
      frame.setIgnoreRepaint(true);
      if (displayMode != null && graphicsDevice.isFullScreenSupported()) {
        frame.setUndecorated(true);
        frame.setLayout(null);

        // Hide the mouse cursor.
        frame.setCursor(Toolkit.getDefaultToolkit().createCustomCursor(
            new BufferedImage(1, 1, BufferedImage.TYPE_4BYTE_ABGR),
            new Point(0, 0), "empty"));

        synchronized(frame.getTreeLock()) {
          graphicsDevice.setFullScreenWindow(frame);
          if (graphicsDevice.isDisplayChangeSupported()) {
            graphicsDevice.setDisplayMode(displayMode);
            isFullScreenMode = true;
            frame.createBufferStrategy(2);
            bufferStrategy = frame.getBufferStrategy();
          } else
            throw new RuntimeException("Display change is not supported.");
        }
      } else {
        Canvas canvas = new Canvas();
        canvas.setSize(width, height);
        frame.add(canvas, BorderLayout.CENTER);
        frame.addWindowListener(DEFAULT_WINDOW_LISTENER);

        // The following code sets the size of the Frame such that the
        // drawing surface is exactly big enough to contain the canvas.
        // The dimensions of the frame border may be obtained using the
        // getInsets() method; however, these dimensions depend on the
        // underlying platform. To obtain valid insets values, the frame must be
        // made displayable by calling pack(). Next, the frame is set to
        // non-resizable before the insets are measured since the border
        // width typically is different in this mode. After the insets are
        // captured, the frame is made resizable and adjusted to contain
        // the canvas. The frame is also positioned to the center of the screen.
        // Finally, the frame is again made non-resizable.
        frame.pack();
        frame.setResizable(false);
        Insets insets = frame.getInsets();
        frame.setResizable(true);
        Point centerPoint = graphicsEnvironment.getCenterPoint();
        int frameWidth = width + insets.left + insets.right;
        int frameHeight = height + insets.top + insets.bottom;
        frame.setBounds(centerPoint.x-frameWidth/2, centerPoint.y-frameHeight/2,
            frameWidth, frameHeight);
        frame.setResizable(false);

        canvas.setIgnoreRepaint(true);
        canvas.createBufferStrategy(2);
        bufferStrategy = canvas.getBufferStrategy();
        frame.setVisible(true);
      }

      BufferCapabilities bufferCapabilities = bufferStrategy.getCapabilities();
      isPageFlipping = bufferCapabilities.isPageFlipping();

      DEFAULT_RENDER_LISTENER.init(isPageFlipping, bufferWidth, bufferHeight);
      GraphicsSource.renderListener = DEFAULT_RENDER_LISTENER;
      clearBufferCount = (isPageFlipping) ? 2 : 1;

      enterDisplayModeSuccessful = true;
    } catch(Throwable e) {
      throw new RuntimeException("Unable to enter display mode.", e);
    }
  }

  /**
   * Of all available display modes with the specified dimensions, it returns
   * the one with the highest bit depth and refresh rate. Bit depth is
   * considered a higher priority.
   * @param width Width of full screen size.
   * @param height Height of full screen size.
   * @return Best display mode or <CODE>null</CODE>
   *   if no such display mode exists in this graphics environment.
   */
  private static DisplayMode getDisplayMode(int width, int height) {
    DisplayMode[] modes = graphicsDevice.getDisplayModes();
    DisplayMode best = null;

    for(int i = modes.length - 1; i >= 0; i--)
      if (modes[i].getWidth() == width && modes[i].getHeight() == height
          && (best == null
              || modes[i].getBitDepth() > best.getBitDepth()
              || (modes[i].getBitDepth() == best.getBitDepth()
                  && modes[i].getRefreshRate() > best.getRefreshRate())))
        best = modes[i];

    return best;
  }

  /**
   * Restores the display mode back to normal and stops the animation.
   */
  public static synchronized void exitDisplayMode() {
    if (frame == null)
      return;

    try {
      stopAnimation();
      frame.dispose();
      isFullScreenMode = false;
      graphicsDevice = null;
      frame = null;
    } catch(Throwable e) {
      throw new RuntimeException("Unable to exit display mode.", e);
    }
  }

  /**
   * Returns whether or not page flipping is used as the double-buffer strategy.
   * Reguardless of the strategy used, the <CODE>RenderListener</CODE> will
   * always be passed a graphics context to the non-visible buffer. After the
   * frame is drawn, it will be displayed in one of two ways. If page flipping
   * is used, then the buffers are instantly swapped by adjusting the video
   * pointer to the opposite buffer; so, even and odd frames are drawn to
   * different buffers. If page flipping is not used, then the contents of the
   * non-visible buffer is copied to the visible one; meaning that drawing
   * always occurs on the same buffer. The value that this method returns is
   * important because it determines how to clean up the buffer to draw the
   * next animation frame.
   * @return <CODE>true</CODE> if page flipping is used as the double-buffer
   *   strategy.
   */
  public static boolean isPageFlipping() {
    checkDisplayMode();
    return isPageFlipping;
  }

  /**
   * Obtains the <CODE>Frame</CODE> used as the full-screen window.
   * @return The full screen window.
   */
  public static Frame getFrame() {
    checkDisplayMode();
    return frame;
  }

  /**
   * Sets the <CODE>ExceptionHandler</CODE> used when an exception is thrown
   * in the animation loop thread. If you pass in <CODE>null</CODE> or if you
   * never call this method, the default exception handler is used. The
   * default exception handler prints the stack trace to standard out
   * and shuts down the application.
   * @param exceptionHandler
   * @see #getExceptionHandler()
   */
  public static void setExceptionHandler(ExceptionHandler exceptionHandler) {
    if (exceptionHandler == null)
      GraphicsSource.exceptionHandler = DEFAULT_EXCEPTION_HANDLER;
    else
      GraphicsSource.exceptionHandler = exceptionHandler;
  }

  /**
   * Obtains the <CODE>ExceptionHandler</CODE> used when an exception is thrown
   * in the animation loop thread.
   * @return Either the <CODE>ExceptionHandler</CODE> passed to
   *   <CODE>setExceptionHandler()</CODE> or the default
   *   <CODE>ExceptionHandler</CODE> if none was ever set.
   * @see #setExceptionHandler(ExceptionHandler)
   */
  public static ExceptionHandler getExceptionHandler() {
    return exceptionHandler;
  }

  /**
   * Obtains the width of the buffers.
   * @return The width of the buffers.
   */
  public static int getBufferWidth() {
    checkDisplayMode();
    return bufferWidth;
  }

  /**
   * Obtains the height of the buffers.
   * @return The height of the buffers.
   */
  public static int getBufferHeight() {
    checkDisplayMode();
    return bufferHeight;
  }

  /**
   * Called at the start of many methods of GraphicsSource to insure that
   * enterDisplayMode() executed.
   * @throws RuntimeException When enterDisplayMode() was never invoked.
   */
  private static void checkDisplayMode() throws RuntimeException {
    if (!enterDisplayModeSuccessful)
      throw new RuntimeException(
          "enterDisplayMode() must be executed prior to this method.");
  }

  /**
   * This constructor is private to prevent instantiation of this class. All
   * methods and fields are static.
   */
  private GraphicsSource() {
  }
}