/**
 * Gaming Tools
 * Copyright (C) 2003  Michael Birken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Michael Birken (o__1@hotmail.com)
 * @version 2.1
 */
package gamingtools;

import java.awt.Graphics;

/**
 * A class that is interested in generating animation frames implements this
 * interface. For each frame, <CODE>GraphicsSource</CODE> invokes
 * <CODE>updateModel()</CODE> to update the state of the game, followed by
 * <CODE>render()</CODE> to display the state to the offscreen image buffer.
 * <CODE>init()</CODE> is invoked prior to frame generation on a different
 * thread than the one for the animation loop. It should be used for loading
 * images and sounds, and for configuration.
 * @see gamingtools.GraphicsSource
 */
public interface RenderListener {

  /**
   * This method is used to load images and sounds, and for configuration. It
   * is invoked prior to frame generation on a different thread than the one
   * for the animation loop.
   * @param isPageFlipping Indicates whether the <CODE>BufferStrategy</CODE>
   *   within <CODE>GraphicsSource</CODE> employs page flipping. If
   *   <CODE>true</CODE>, then even frames and odd frames are drawn to different
   *   offscreen buffers. Otherwise, all frames are drawn to the same buffer.
   *   This parameter determines how the buffer must be updated to generate the
   *   next frame.
   * @param bufferWidth The width of the buffers.
   * @param bufferHeight The height of the bufrers.
   */
  public void init(boolean isPageFlipping, int bufferWidth, int bufferHeight);

  /**
   * This method is used to update the state of the game.
   * <CODE>GraphicsSource</CODE> tries to sustain the apparent frame rate by
   * assuming that the time required to update the game state is significantly
   * less than the time to display it. It will occasionally and typically
   * unnoticeably invoke <CODE>updateModel()</CODE> more often than
   * <CODE>render()</CODE>, to effectively skips frames.
   */
  public void updateModel();

  /**
   * Displays the game state to the back buffer.
   * @param g The graphics context used to draw the animation frame. Do not
   *   invoke <CODE>g.dispose()</CODE>. Disposal of this graphics context is
   *   automatically performed by <CODE>GraphicsSource</CODE>.
   * @param isBufferCleared Indicates whether the buffer is in the same
   *   state since the last time it was drawn to. Occationally, the buffer
   *   may be losted and restored. When this happens, it is effectively erased.
   *   If this parameter is <CODE>true</CODE>, then the entire buffer needs to
   *   be painted. Otherwise, only the portion of the buffer that changes for
   *   this frame needs to be updated. Note, after calling
   *   <CODE>GraphicsSource.setRenderListener()</CODE>, this parameter will
   *   be <CODE>true</CODE> for the first frame or two depending upon which
   *   buffer strategy is employed.
   * @param bufferIndex If page flipping is used, then this parameter will
   *   alternate between 0 and 1. Otherwise, it will always be 0.
   */
  public void render(Graphics g, boolean isBufferCleared, int bufferIndex);
}