/**
 * Gaming Tools
 * Copyright (C) 2003  Michael Birken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Michael Birken (o__1@hotmail.com)
 * @version 2.1
 */
package gamingtools;

import java.net.*;
import javax.sound.sampled.*;

/**
 * The <CODE>SoundEffect</CODE> class provides methods for loading and playing
 * a sound effect. When this object is no longer needed, the
 * <CODE>dispose()</CODE> method should be invoked.
 */
class SoundEffect implements Audio {

  /** Loaded audio data. */
  private Clip clip;
  /** Used to determine when the sound effect finishes playing. */
  private LineListener lineListener;

  /**
   * Constructs a <CODE>SoundEffect</CODE> out of an audio file.
   * @param fileURL Supported file formats are AIFF-C, AIFF, AU, SND, and WAVE.
   */
  public SoundEffect(URL fileURL) {
    this(fileURL, null);
  }

  /**
   * Constructs a <CODE>SoundEffect</CODE> out of an audio file and associates
   * an <CODE>AudioListener</CODE> with it.
   * @param fileURL Supported file formats are AIFF-C, AIFF, AU, SND, and WAVE.
   * @param audioListener Called back when the audio stops playing.
   */
  public SoundEffect(URL fileURL, final AudioListener audioListener) {

    try {
      AudioInputStream audioInputStream
          = AudioSystem.getAudioInputStream(fileURL);
      AudioFormat format = audioInputStream.getFormat();
      DataLine.Info info = new DataLine.Info(
          Clip.class, format, AudioSystem.NOT_SPECIFIED);
      clip = (Clip)AudioSystem.getLine(info);
      setAudioListener(audioListener);
      clip.open(audioInputStream);
    } catch(Throwable t) {
      throw new RuntimeException(
          "Error loading sound effect \"" + fileURL + "\".", t);
    }
  }

  /**
   * Starts playing this sound effect. Each time this method is called, the
   * sound effect is restarted from the beginning.
   */
  public void play() {
    if (clip != null) {
      stop();
      clip.setFramePosition(0);
      clip.start();
    }
  }

  /**
   * Starts playing this sound effect in a loop.
   */
  public void loop() {
    if (clip != null) {
      stop();
      clip.loop(Clip.LOOP_CONTINUOUSLY);
    }
  }

  /**
   * Stops playing this sound effect.
   */
  public void stop() {
    if (clip != null && clip.isRunning())
      clip.stop();
  }

  /**
   * Registers an <CODE>AudioListener</CODE> with this <CODE>SoundEffect</CODE>
   * object to receive notification of when the sound effect stops playing.
   * @param audioListener Called back when the sound effect stops playing.
   */
  public void setAudioListener(final AudioListener audioListener) {
    clearAudioListener();
    if (audioListener != null) {
      lineListener = new LineListener() {
        public void update(LineEvent event) {
          if (event.getType() == LineEvent.Type.STOP)
            audioListener.audioStopped();
        }
      };
      clip.addLineListener(lineListener);
    }
  }

  /**
   * Unregisters the <CODE>AudioListener</CODE> if there is one. This method is
   * equivalent to passing <CODE>null</CODE> to <CODE>setAudioListener()</CODE>.
   */
  public void clearAudioListener() {
    if (clip != null && lineListener != null)
      clip.removeLineListener(lineListener);
    lineListener = null;
  }

  /**
   * <P>Releases any system resources used by this object. A
   * <CODE>SoundEffect</CODE> object cannot be used after calling this
   * method.</P>
   * <P>The finalization process of the garbage collector also disposes of
   * the same system resources; however, since the finalization process may
   * not run to completion for a long duration, it is preferable to manually
   * free the resources by invoking this method.</P>
   */
  public void dispose() {
    if (clip != null) {
      stop();
      clearAudioListener();
      if (clip.isOpen())
        clip.close();
      lineListener = null;
      clip = null;
    }
  }

  /**
   * Invokes <CODE>dispose()</CODE> during the finalization process of the
   * garbage collector.
   */
  protected void finalize() {
    dispose();
  }
}
